import sys,string,re

import wx
import wx.stc as stc

from Script import Script
import globals

class ScriptEditor(wx.SplitterWindow):
    def __init__(self,parent,id):
        wx.SplitterWindow.__init__(self,parent,id)
        splitter = self
        box = wx.BoxSizer(wx.VERTICAL)
        win = wx.Panel(splitter,-1)
        self.notebook = wx.Notebook(win,-1)
        buttonPanel = wx.Panel(win,-1)
        box.Add(buttonPanel,0,wx.ALIGN_TOP)
        box.Add(self.notebook,1,wx.EXPAND)
        box2 = wx.BoxSizer(wx.HORIZONTAL)
        self.compileButton = wx.Button(buttonPanel,-1,"Compile")
        box2.Add(self.compileButton,0,wx.EXPAND)
        self.closeButton = wx.Button(buttonPanel,-1,"Close Script Tab")
        box2.Add(self.closeButton,0,wx.EXPAND)
        buttonPanel.SetSizerAndFit(box2)
        self.compileButton.Bind(wx.EVT_BUTTON,self.OnCompile)
        self.compileButton.Enable(False)
        self.closeButton.Bind(wx.EVT_BUTTON,self.OnCloseEditor)
        self.closeButton.Enable(False)
        self.output = wx.ListCtrl(splitter,-1,style=wx.LC_REPORT|wx.LC_NO_HEADER)
        self.output.InsertColumn(0,"Output")
        self.output.SetColumnWidth(0,parent.GetSize().width)
#        self.output.SetEditable(False)
        self.output.Bind(wx.EVT_LIST_ITEM_SELECTED,self.OnOutSelected)
        win.SetSizerAndFit(box)

        splitter.SplitHorizontally(win,self.output,-100)
        splitter.SetMinimumPaneSize(70)
        
        
        self.scripts = []
        self.scriptTable = {}
        self.editors = []
        
        self.lastHeight = self.GetSize().height

        self.Bind(wx.EVT_SIZE,self.OnSize)

    def OnCompile(self,event):
        self.compile()

    def OnCloseEditor(self,event):
        self.removeCurrentTab()
        
    def OnSize(self,event):
        pos = self.GetSashPosition()
        if pos < 0:
            self.lastHeight = self.GetSize().height
            event.Skip()
            return        
        self.SetSashPosition(-(self.lastHeight-pos) )
        self.lastHeight = self.GetSize().height
        self.output.SetColumnWidth(0,self.GetSize().width)

    def OnOutSelected(self,event):        
        line = event.GetIndex()
        r = re.compile('\([0-9]+\)')
        num = r.findall(self.output.GetItemText(line))
        if num:
            editor = self.getCurrentEditor()
            row = int(num[0][1:-1])-1
            editor.EnsureVisible(row)
            editor.GotoLine(row)            
            editor.SetSelection(editor.GetCurrentPos(),editor.GetLineEndPosition(row))
            
    def getCurrentEditor(self):
        return self.editors[self.notebook.GetSelection()]
    
    def compile(self):
        script = self.getCurrentScript()
        script.setUnixData(self.editors[self.notebook.GetSelection()].GetText())
        compiled,err = script.compile()
        prefix = "Errors occured during compile:\n"
        if compiled:
            prefix = "Compile Successful!\n"
        if err:
            prefix += err
        self.output.DeleteAllItems()
        lines = prefix.split('\n')
        for i,line in enumerate(lines):
            if line:
                self.output.InsertStringItem(i,line)
        return compiled

    def getCurrentScript(self):
        return self.scripts[self.notebook.GetSelection()]
    
    def getScript(self,name):
        s = self.scriptTable.get(name,None)
        if s:
            return s[0]

    def selectScript(self,name):
        s = self.scriptTable.get(name,None)
        if s:
            self.notebook.SetSelection(s[1])
            return True
        return False
            
    def addScript(self,s):
        if not self.selectScript(s.getName()):
            self.scripts.append(s)
            editor = self.makeEditor()
            editor.SetText(s.getUnixData())
            self.editors.append(editor)
            self.notebook.AddPage(editor,s.getName()[:-4])
            self.notebook.SetSelection(len(self.scripts)-1)
            self.scriptTable[s.getName()] = (s,len(self.scripts)-1)
        self.compileButton.Enable(True)
        self.closeButton.Enable(True)

    def removeCurrentTab(self):
        self.compile()
        index = self.notebook.GetSelection()
        s = self.getCurrentScript()
        self.scripts = self.scripts[:index] + self.scripts[index+1:]
        self.editors = self.editors[:index] + self.editors[index+1:]
        del self.scriptTable[s.getName()]
        self.notebook.DeletePage(index)
        if not self.notebook.GetPageCount():
            self.closeButton.Enable(False)
            self.compileButton.Enable(False)
        
    def makeEditor(self):
        textCtl = stc.StyledTextCtrl(self.notebook,-1)
        textCtl.SetLexerLanguage('cpp')
        textCtl.SetKeyWords(0,Script.lang_keywords)
        textCtl.SetKeyWords(1,string.join(Script.nwscript_keywords.keys()))

        textCtl.SetProperty("fold","1")
        
        textCtl.SetViewWhiteSpace(False)
        textCtl.SetUseAntiAliasing(True)
        
        #textCtl.SetEdgeMode(stc.STC_EDGE_LINE)
        #textCtl.SetEdgeColumn(78)

        # Setup a margin to hold fold markers
        textCtl.SetMarginType(2, stc.STC_MARGIN_SYMBOL)
        textCtl.SetMarginMask(2, stc.STC_MASK_FOLDERS)
        textCtl.SetMarginSensitive(2, True)
        textCtl.SetMarginWidth(2, 12)
        
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDEROPEN,
                          stc.STC_MARK_CIRCLEMINUS,          "white", "#404040");
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDER,
                          stc.STC_MARK_CIRCLEPLUS,           "white", "#404040");
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDERSUB,
                          stc.STC_MARK_VLINE,                "white", "#404040");
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDERTAIL,
                          stc.STC_MARK_LCORNERCURVE,         "white", "#404040");
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDEREND,
                          stc.STC_MARK_CIRCLEPLUSCONNECTED,  "white", "#404040");
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDEROPENMID,
                          stc.STC_MARK_CIRCLEMINUSCONNECTED, "white", "#404040");
        textCtl.MarkerDefine(stc.STC_MARKNUM_FOLDERMIDTAIL,
                          stc.STC_MARK_TCORNERCURVE,         "white", "#404040");
        
        faces = None
        if wx.Platform == '__WXMSW__':
            faces = { 'times': 'Times New Roman',
                      'mono' : 'Courier New',
                      'helv' : 'Arial',
                      'other': 'Comic Sans MS',
                      'size' : 10,
                      'size2': 8,
                      }
        else:
            faces = { 'times': 'Times',
                      'mono' : 'Courier',
                      'helv' : 'Helvetica',
                      'other': 'new century schoolbook',
                      'size' : 12,
                      'size2': 10,
                      }

        textCtl.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % faces)
        textCtl.StyleClearAll()  # Reset all to be like the default

        # Global default styles for all languages
        textCtl.StyleSetSpec(stc.STC_STYLE_DEFAULT,     "face:%(helv)s,size:%(size)d" % faces)
        textCtl.StyleSetSpec(stc.STC_STYLE_LINENUMBER,  "back:#C0C0C0,face:%(helv)s,size:%(size2)d" % faces)
        textCtl.StyleSetSpec(stc.STC_STYLE_CONTROLCHAR, "face:%(other)s" % faces)
        textCtl.StyleSetSpec(stc.STC_STYLE_BRACELIGHT,  "fore:#FFFFFF,back:#0000FF,bold")
        textCtl.StyleSetSpec(stc.STC_STYLE_BRACEBAD,    "fore:#000000,back:#FF0000,bold")

        # C styles
        # Default 
        textCtl.StyleSetSpec(stc.STC_C_DEFAULT, "fore:#000000,face:%(helv)s,size:%(size)d" % faces)
        # Comments
        textCtl.StyleSetSpec(stc.STC_C_COMMENTLINE, "fore:#007F00,face:%(other)s,size:%(size)d" % faces)
        # Number
        textCtl.StyleSetSpec(stc.STC_C_NUMBER, "fore:#007F7F,size:%(size)d" % faces)
        # String
        textCtl.StyleSetSpec(stc.STC_C_STRING, "fore:#7F007F,face:%(helv)s,size:%(size)d" % faces)
        # Single quoted string
        textCtl.StyleSetSpec(stc.STC_C_CHARACTER, "fore:#7F007F,face:%(helv)s,size:%(size)d" % faces)
        # Keyword
        textCtl.StyleSetSpec(stc.STC_C_WORD, "fore:#00007F,bold,size:%(size)d" % faces)
        # Keyword2
        textCtl.StyleSetSpec(stc.STC_C_WORD2, "fore:#007F00,size:%(size)d" % faces)
        # Operators
        textCtl.StyleSetSpec(stc.STC_C_OPERATOR, "bold,size:%(size)d" % faces)
        # Identifiers
        textCtl.StyleSetSpec(stc.STC_C_IDENTIFIER, "fore:#000000,face:%(helv)s,size:%(size)d" % faces)
        # Comment-blocks
        textCtl.StyleSetSpec(stc.STC_C_COMMENT, "fore:#7F7F7F,size:%(size)d" % faces)
        # End of line where string is not closed
        textCtl.StyleSetSpec(stc.STC_C_STRINGEOL,
                             "fore:#000000,face:%(mono)s,back:#E0C0E0,eol,size:%(size)d" % faces)

        textCtl.SetCaretForeground("BLUE")
        textCtl.SetMarginType(1, stc.STC_MARGIN_NUMBER)
        textCtl.SetMarginWidth(1, 25)

        textCtl.Bind(stc.EVT_STC_UPDATEUI, self.OnUpdateUI)
        textCtl.Bind(stc.EVT_STC_MARGINCLICK, self.OnMarginClick)
        #self.Bind(wx.EVT_KEY_DOWN, self.OnKeyPressed)

        return textCtl
    
    def OnUpdateUI(self, evt):
        editor = evt.GetEventObject()
        # check for matching braces
        braceAtCaret = -1
        braceOpposite = -1
        charBefore = None
        caretPos = editor.GetCurrentPos()

        if caretPos > 0:
            charBefore = editor.GetCharAt(caretPos - 1)
            styleBefore = editor.GetStyleAt(caretPos - 1)

        # check before
        if charBefore and chr(charBefore) in "[]{}()" and styleBefore == stc.STC_P_OPERATOR:
            braceAtCaret = caretPos - 1

        # check after
        if braceAtCaret < 0:
            charAfter = editor.GetCharAt(caretPos)
            styleAfter = editor.GetStyleAt(caretPos)

            if charAfter and chr(charAfter) in "[]{}()" and styleAfter == stc.STC_P_OPERATOR:
                braceAtCaret = caretPos

        if braceAtCaret >= 0:
            braceOpposite = editor.BraceMatch(braceAtCaret)

        if braceAtCaret != -1  and braceOpposite == -1:
            editor.BraceBadLight(braceAtCaret)
        else:
            editor.BraceHighlight(braceAtCaret, braceOpposite)

    def OnMarginClick(self, evt):
        editor = evt.GetEventObject()
        # fold and unfold as needed
        if evt.GetMargin() == 2:
            if evt.GetShift() and evt.GetControl():
                self.FoldAll(editor)
            else:
                lineClicked = editor.LineFromPosition(evt.GetPosition())

                if editor.GetFoldLevel(lineClicked) & stc.STC_FOLDLEVELHEADERFLAG:
                    if evt.GetShift():
                        editor.SetFoldExpanded(lineClicked, True)
                        self.Expand(editor,lineClicked, True, True, 1)
                    elif evt.GetControl():
                        if editor.GetFoldExpanded(lineClicked):
                            editor.SetFoldExpanded(lineClicked, False)
                            self.Expand(editor,lineClicked, False, True, 0)
                        else:
                            editor.SetFoldExpanded(lineClicked, True)
                            self.Expand(editor,lineClicked, True, True, 100)
                    else:
                        editor.ToggleFold(lineClicked)
                        
    def FoldAll(self,editor):
        lineCount = editor.GetLineCount()
        expanding = True

        # find out if we are folding or unfolding
        for lineNum in range(lineCount):
            if editor.GetFoldLevel(lineNum) & stc.STC_FOLDLEVELHEADERFLAG:
                expanding = not editor.GetFoldExpanded(lineNum)
                break;

        lineNum = 0

        while lineNum < lineCount:
            level = editor.GetFoldLevel(lineNum)
            if level & stc.STC_FOLDLEVELHEADERFLAG and \
               (level & stc.STC_FOLDLEVELNUMBERMASK) == stc.STC_FOLDLEVELBASE:

                if expanding:
                    editor.SetFoldExpanded(lineNum, True)
                    lineNum = self.Expand(editor,lineNum, True)
                    lineNum = lineNum - 1
                else:
                    lastChild = editor.GetLastChild(lineNum, -1)
                    editor.SetFoldExpanded(lineNum, False)

                    if lastChild > lineNum:
                        editor.HideLines(lineNum+1, lastChild)

            lineNum = lineNum + 1

    def Expand(self, editor, line, doExpand, force=False, visLevels=0, level=-1):
        lastChild = editor.GetLastChild(line, level)
        line = line + 1

        while line <= lastChild:
            if force:
                if visLevels > 0:
                    editor.ShowLines(line, line)
                else:
                    editor.HideLines(line, line)
            else:
                if doExpand:
                    editor.ShowLines(line, line)

            if level == -1:
                level = editor.GetFoldLevel(line)

            if level & stc.STC_FOLDLEVELHEADERFLAG:
                if force:
                    if visLevels > 1:
                        editor.SetFoldExpanded(line, True)
                    else:
                        editor.SetFoldExpanded(line, False)

                    line = editor.Expand(line, doExpand, force, visLevels-1)

                else:
                    if doExpand and editor.GetFoldExpanded(line):
                        line = editor.Expand(line, True, force, visLevels-1)
                    else:
                        line = editor.Expand(line, False, force, visLevels-1)
            else:
                line = line + 1;

        return line
    
if __name__ == '__main__':
    win = None
    class MyApp(wx.App):
        def OnInit(self):
            global win
            import ResourceManager
            globals.resourceManager = ResourceManager.ResourceManager()
            if sys.platform == 'darwin':
                globals.resourceManager.scanGameDir('/Applications/Neverwinter Nights/')
            else:
                globals.resourceManager.scanGameDir('/home/pgorniak/nwn/')                
            Script.init_nwscript_keywords()
            frame = wx.Frame(None,-1,"Script Editor",wx.DefaultPosition,wx.Size(600,400))
            win = ScriptEditor(frame,-1)
            win.SetFocus()
            self.SetTopWindow(frame)
            frame.Show(True)
            frame.SetSize((601,400))
            return True
    app = MyApp(0)
    s = Script(sys.argv[1],open(sys.argv[1]).read())
    s.setNWNDir('/Applications/Neverwinter Nights')
    win.addScript(s)
    app.MainLoop()
